import pytest
from typing import *
from dataclasses import dataclass
from dataclasses_json import dataclass_json
from marshmallow import ValidationError


# == Common use cases ==
@dataclass_json
@dataclass
class C1:
    f1: Union[int, str]


@dataclass_json
@dataclass
class C2:
    f1: Union[int, Dict[str, float]]


@dataclass_json
@dataclass
class C3:
    f1: Union[int, List[float]]


# == Use cases with nested dataclasses ==
@dataclass_json
@dataclass
class Aux1:
    f1: int


@dataclass_json
@dataclass
class Aux2:
    f1: str


@dataclass_json
@dataclass
class Aux3:
    f2: str


@dataclass_json
@dataclass
class C4:
    f1: Union[Aux1, Aux2]


@dataclass_json
@dataclass
class C12:
    f1: Union[Aux2, Aux3]


@dataclass_json
@dataclass
class C5:
    f1: Union[Aux1, Aux2, None]


@dataclass_json
@dataclass
class C6:
    f1: Union[Aux1, None]  # The same as Optional[Aux1]


@dataclass_json
@dataclass
class C7:
    f1: Union[C5, C6]


@dataclass_json
@dataclass
class C8:
    f1: Dict[str, Union[Aux1, Aux2]]


@dataclass_json
@dataclass
class C9:
    f1: List[Union[Aux1, Aux2]]


try:
    @dataclass_json
    @dataclass
    class C10:
        """
         Use py3.10+ pipe notation for optionals
         Check if passing None as the first type option doesn't break type resolution

         Instantiate with None default so NoneType is returned for runtime field value type.
        """
        f1: None | str = None
except TypeError:
    @dataclass_json
    @dataclass
    class C10:
        """
         Replace test case on versions prior to 3.10
        """
        f1: Union[None, str] = None


try:
    @dataclass_json
    @dataclass
    class C11:
        """
         Use py3.10+ pipe notation for optionals
         Check if passing None as the second type option doesn't break type resolution

         Instantiate with None default so NoneType is returned for runtime field value type.
        """
        f1: str | None = None
except TypeError:
    @dataclass_json
    @dataclass
    class C11:
        """
         Replace test case on versions prior to 3.10
        """
        f1: Union[str, None] = None


@dataclass_json
@dataclass
class C13:
    data: str


@dataclass_json
@dataclass
class C14:
    content: str


@dataclass_json
@dataclass
class C15:
    data: C13


@dataclass_json
@dataclass
class C16:
    event: Union[C15, C13]


params = [
    (C1(f1=12), {"f1": 12}, '{"f1": 12}'),
    (C1(f1="str1"), {"f1": "str1"}, '{"f1": "str1"}'),

    (C2(f1=10), {"f1": 10}, '{"f1": 10}'),
    (C2(f1={"str1": 0.12}), {"f1": {"str1": 0.12}}, '{"f1": {"str1": 0.12}}'),

    (C3(f1=10), {"f1": 10}, '{"f1": 10}'),
    (C3(f1=[0.12, 0.13, 0.14]), {"f1": [0.12, 0.13, 0.14]}, '{"f1": [0.12, 0.13, 0.14]}'),

    (C4(f1=Aux1(1)), {"f1": {"f1": 1, "__type": "Aux1"}}, '{"f1": {"f1": 1, "__type": "Aux1"}}'),
    (C4(f1=Aux2("str1")), {"f1": {"f1": "str1", "__type": "Aux2"}}, '{"f1": {"f1": "str1", "__type": "Aux2"}}'),

    (C5(f1=Aux1(1)), {"f1": {"f1": 1, "__type": "Aux1"}}, '{"f1": {"f1": 1, "__type": "Aux1"}}'),
    (C5(f1=Aux2("str1")), {"f1": {"f1": "str1", "__type": "Aux2"}}, '{"f1": {"f1": "str1", "__type": "Aux2"}}'),
    (C5(f1=None), {"f1": None}, '{"f1": null}'),

    (C6(f1=Aux1(1)), {"f1": {"f1": 1}}, '{"f1": {"f1": 1}}'),  # For Optionals, type can be clearly defined
    (C6(f1=None), {"f1": None}, '{"f1": null}'),

    (C7(C5(Aux2("str1"))),
     {"f1": {"f1": {"f1": "str1", "__type": "Aux2"}, "__type": "C5"}},
     '{"f1": {"f1": {"f1": "str1", "__type": "Aux2"}, "__type": "C5"}}'),
    (C7(C6(Aux1(12))),
     {"f1": {"f1": {"f1": 12}, "__type": "C6"}},
     '{"f1": {"f1": {"f1": 12}, "__type": "C6"}}'),

    (C8({"str1": Aux1(12), "str2": Aux2("str3")}),
     {"f1": {"str1": {"f1": 12, "__type": "Aux1"}, "str2": {"f1": "str3", "__type": "Aux2"}}},
     '{"f1": {"str1": {"f1": 12, "__type": "Aux1"}, "str2": {"f1": "str3", "__type": "Aux2"}}}'),

    (C9([Aux1(12), Aux2("str3")]),
     {"f1": [{"f1": 12, "__type": "Aux1"}, {"f1": "str3", "__type": "Aux2"}]},
     '{"f1": [{"f1": 12, "__type": "Aux1"}, {"f1": "str3", "__type": "Aux2"}]}'),

    (C10(f1=None), {"f1": None}, '{"f1": null}'),
    (C10(f1='str1'), {"f1": 'str1'}, '{"f1": "str1"}'),

    (C11(f1=None), {"f1": None}, '{"f1": null}')
]


@pytest.mark.parametrize('obj, expected, expected_json', params)
def test_serialize(obj, expected, expected_json):
    s = obj.schema()
    assert s.dump(obj) == expected
    assert s.dumps(obj) == expected_json


@pytest.mark.parametrize('expected_obj, data, data_json', params)
def test_deserialize(expected_obj, data, data_json):
    cls = type(expected_obj)
    s = cls.schema()
    assert s.load(data) == expected_obj
    assert s.loads(data_json) == expected_obj


def test_deserialize_twice():
    data = {"f1": [{"f1": 12, "__type": "Aux1"}, {"f1": "str3", "__type": "Aux2"}]}
    expected_obj = C9([Aux1(12), Aux2("str3")])

    s = C9.schema()
    res1 = s.load(data)
    res2 = s.load(data)
    assert res1 == expected_obj and res2 == expected_obj


@pytest.mark.parametrize('obj', [
    (C2(f1={"str1": "str1"})),
    (C3(f1=[0.12, 0.13, "str1"])),
])
def test_serialize_with_error(obj):
    s = obj.schema()
    with pytest.raises(ValueError):
        assert s.dump(obj)


@pytest.mark.parametrize('cls, data', [
    (C1, {"f1": None}),
])
def test_deserialize_with_error(cls, data):
    s = cls.schema()
    with pytest.raises(ValidationError):
        assert s.load(data)


def test_deserialize_without_discriminator():
    # determine based on type
    json = '{"f1": {"f1": 1}}'
    s = C4.schema()
    obj = s.loads(json)
    assert obj.f1 is not None
    assert type(obj.f1) == Aux1

    json = '{"f1": {"f1": "str1"}}'
    s = C4.schema()
    obj = s.loads(json)
    assert obj.f1 is not None
    assert type(obj.f1) == Aux2

    # determine based on field name
    json = '{"f1": {"f1": "str1"}}'
    s = C12.schema()
    obj = s.loads(json)
    assert obj.f1 is not None
    assert type(obj.f1) == Aux2
    json = '{"f1": {"f2": "str1"}}'
    s = C12.schema()
    obj = s.loads(json)
    assert obj.f1 is not None
    assert type(obj.f1) == Aux3

    # if no matching types, type should remain dict
    json = '{"f1": {"f3": "str2"}}'
    s = C12.schema()
    obj = s.loads(json)
    assert type(obj.f1) == dict


def test_deserialize_with_mismatched_field_types():
    json = '{"event": {"data": "Hello world!"} }'
    s = C16.schema()
    obj = s.loads(json)
    assert obj.event is not None
    assert obj.event.data == "Hello world!"



import dataclasses
import json
from enum import Enum
from typing import Tuple

from dataclasses import dataclass

from dataclasses_json import dataclass_json


class MyEnum(Enum):
    STR1 = "str1"
    STR2 = "str2"
    STR3 = "str3"
    INT1 = 1
    FLOAT1 = 1.23


@dataclass_json
@dataclass
class DataWithHeterogeneousTuple:
    my_tuple: Tuple[MyEnum, str, float]


hetero_tuple_json = '{"my_tuple": ["str1", "str2", 1.23]}'
hetero_tuple_data = DataWithHeterogeneousTuple(my_tuple=(MyEnum.STR1, "str2", 1.23))


@dataclass_json
@dataclass
class DataWithNestedTuple:
    my_tuple: Tuple[Tuple[int, str], Tuple[MyEnum, int], int]


nested_tuple_json = '{"my_tuple": [[1, "str1"], ["str2", 1], 1]}'
nested_tuple_data = DataWithNestedTuple(my_tuple=((1, "str1"), (MyEnum.STR2, 1), 1))


@dataclass_json
@dataclass
class DataWithEllipsisTuple:
    my_tuple: Tuple[str, ...]


ellipsis_tuple_json = '{"my_tuple": ["str1", "str2", "str3"]}'
ellipsis_tuple_data = DataWithEllipsisTuple(my_tuple=("str1", "str2", "str3"))


class TestEncoder:
    def test_enum_with_tuple(self):
        assert hetero_tuple_data.to_json() == hetero_tuple_json, f'Actual: {hetero_tuple_data.to_json()}, Expected: {hetero_tuple_json}'
        assert hetero_tuple_data.to_dict(encode_json=True) == json.loads(hetero_tuple_json), f'Actual: {hetero_tuple_data.to_dict()}, Expected: {json.loads(hetero_tuple_json)}'
    
    def test_nested_tuple(self):
        assert nested_tuple_data.to_json() == nested_tuple_json, f'Actual: {nested_tuple_data.to_json()}, Expected: {nested_tuple_json}'
        assert nested_tuple_data.to_dict(encode_json=True) == json.loads(nested_tuple_json), f'Actual: {nested_tuple_data.to_dict()}, Expected: {json.loads(nested_tuple_json)}'
    
    def test_ellipsis_tuple(self):
        assert ellipsis_tuple_data.to_json() == ellipsis_tuple_json, f'Actual: {ellipsis_tuple_data.to_json()}, Expected: {ellipsis_tuple_json}'
        assert ellipsis_tuple_data.to_dict(encode_json=True) == json.loads(ellipsis_tuple_json), f'Actual: {ellipsis_tuple_data.to_dict()}, Expected: {json.loads(ellipsis_tuple_json)}'


class TestSchemaEncoder:
    def test_enum_with_tuple(self):
        js = DataWithHeterogeneousTuple.schema().dumps(hetero_tuple_data)
        assert js == hetero_tuple_json, f'Actual: {js}, Expected: {hetero_tuple_json}'

    def test_nested_tuple(self):
        js = DataWithNestedTuple.schema().dumps(nested_tuple_data)
        assert js == nested_tuple_json, f'Actual: {js}, Expected: {nested_tuple_json}'

    def test_ellipsis_tuple(self):
        js = DataWithEllipsisTuple.schema().dumps(ellipsis_tuple_data)
        assert js == ellipsis_tuple_json, f'Actual: {js}, Expected: {ellipsis_tuple_json}'
        
    
    
class TestDecoder:
    def test_enum_with_tuple(self):
        tuple_data_from_json = DataWithHeterogeneousTuple.from_json(hetero_tuple_json)
        assert hetero_tuple_data == tuple_data_from_json
        assert tuple_data_from_json.to_json() == hetero_tuple_json

    def test_nested_tuple(self):
        tuple_data_from_json = DataWithNestedTuple.from_json(nested_tuple_json)
        assert nested_tuple_data == tuple_data_from_json
        assert tuple_data_from_json.to_json() == nested_tuple_json

    def test_ellipsis_tuple(self):
        tuple_data_from_json = DataWithEllipsisTuple.from_json(ellipsis_tuple_json)
        assert ellipsis_tuple_data == tuple_data_from_json
        assert tuple_data_from_json.to_json() == ellipsis_tuple_json


class TestSchemaDecoder:
    def test_enum_with_tuple(self):
        tuple_data_from_json = DataWithHeterogeneousTuple.schema().loads(hetero_tuple_json)
        assert hetero_tuple_data == tuple_data_from_json
        assert tuple_data_from_json.to_json() == hetero_tuple_json

    def test_nested_tuple(self):
        tuple_data_from_json = DataWithNestedTuple.schema().loads(nested_tuple_json)
        assert nested_tuple_data == tuple_data_from_json
        assert tuple_data_from_json.to_json() == nested_tuple_json

    def test_ellipsis_tuple(self):
        tuple_data_from_json = DataWithEllipsisTuple.schema().loads(ellipsis_tuple_json)
        assert ellipsis_tuple_data == tuple_data_from_json
        assert tuple_data_from_json.to_json() == ellipsis_tuple_json



from dataclasses import dataclass
from typing import Optional

from dataclasses_json import DataClassJsonMixin


@dataclass(frozen=True)
class Tree(DataClassJsonMixin):
    value: str
    left: Optional['Tree']
    right: Optional['Tree']


family_tree_json = """
{
    "value": "Boy",
    "left": {
        "value": "Ma",
        "left": {
            "value": "Maternal Grandma",
            "left": null,
            "right": null
        },
        "right": {
            "value": "Maternal Grandpa",
            "left": null,
            "right": null
        }
    },
    "right": {
        "value": "Pa",
        "left": {
            "value": "Paternal Grandma",
            "left": null,
            "right": null
        },
        "right": {
            "value": "Paternal Grandpa",
            "left": null,
            "right": null
        }
    }
}
""".strip()

family_tree = Tree(
    "Boy",
    Tree(
        "Ma",
        Tree(
            "Maternal Grandma",
            None,
            None
        ),
        Tree(
            "Maternal Grandpa",
            None,
            None
        )
    ),
    Tree("Pa",
         Tree(
             "Paternal Grandma",
             None,
             None
         ),
         Tree(
             "Paternal Grandpa",
             None,
             None
         )
         )
)


class TestRecursive:
    def test_tree_encode(self):
        assert family_tree.to_json(indent=4) == family_tree_json

    def test_tree_decode(self):
        assert Tree.from_json(family_tree_json) == family_tree

    def test_tree_schema_round_trip(self):
        tree_dict = Tree.schema().dump(family_tree)
        tree_obj = Tree.schema().load(tree_dict)
        assert tree_obj == family_tree




import sys
from collections import deque
from dataclasses import dataclass, field
from datetime import datetime
from decimal import Decimal
from typing import (Collection,
                    Deque,
                    Dict,
                    FrozenSet,
                    List,
                    NewType,
                    Optional,
                    Set,
                    Tuple,
                    TypeVar,
                    Union,
                    Any)
from uuid import UUID
if sys.version_info >= (3, 9):
    from collections import Counter
else:
    from typing import Counter

from marshmallow import fields

from dataclasses_json import (DataClassJsonMixin, LetterCase, dataclass_json)
from dataclasses_json.cfg import config

A = TypeVar('A')
UUIDWrapper = NewType('UUIDWrapper', UUID)
UUIDWrapperWrapper = NewType('UUIDWrapperWrapper', UUIDWrapper)


@dataclass(frozen=True)
class DataClassWithDecimal(DataClassJsonMixin):
    x: Decimal


@dataclass(frozen=True)
class DataClassWithNewType(DataClassJsonMixin):
    id: UUIDWrapper


@dataclass(frozen=True)
class DataClassWithNestedNewType(DataClassJsonMixin):
    id: UUIDWrapperWrapper


@dataclass(frozen=True)
class DataClassWithList(DataClassJsonMixin):
    xs: List[int]


@dataclass(frozen=True)
class DataClassWithListBuiltin(DataClassJsonMixin):
    xs: list


@dataclass(frozen=True)
class DataClassWithListUnbound(DataClassJsonMixin):
    xs: List


@dataclass(frozen=True)
class DataClassWithListDefaultFactory(DataClassJsonMixin):
    xs: List[int] = field(default_factory=list)


@dataclass(frozen=True)
class DataClassWithListStr(DataClassJsonMixin):
    xs: List[str]


@dataclass(frozen=True)
class DataClassWithDict(DataClassJsonMixin):
    kvs: Dict[str, str]


@dataclass(frozen=True)
class DataClassWithDictBuiltin(DataClassJsonMixin):
    kvs: Dict


@dataclass(frozen=True)
class DataClassWithDictUnbound(DataClassJsonMixin):
    kvs: Dict


@dataclass(frozen=True)
class DataClassWithDictInt(DataClassJsonMixin):
    kvs: Dict[int, str]


@dataclass(frozen=True)
class DataClassWithDictDefaultFactory(DataClassJsonMixin):
    kvs: Dict[str, str] = field(default_factory=dict)


@dataclass(frozen=True)
class DataClassWithSet(DataClassJsonMixin):
    xs: Set[int]


@dataclass(frozen=True)
class DataClassWithSetBuiltin(DataClassJsonMixin):
    xs: set


@dataclass(frozen=True)
class DataClassWithSetUnbound(DataClassJsonMixin):
    xs: Set


@dataclass(frozen=True)
class DataClassWithTuple(DataClassJsonMixin):
    xs: Tuple[int]


@dataclass(frozen=True)
class DataClassWithTupleBuiltin(DataClassJsonMixin):
    xs: tuple


@dataclass(frozen=True)
class DataClassWithTupleUnbound(DataClassJsonMixin):
    xs: Tuple


@dataclass(frozen=True)
class DataClassWithFrozenSet(DataClassJsonMixin):
    xs: FrozenSet[int]


@dataclass(frozen=True)
class DataClassWithFrozenSetBuiltin(DataClassJsonMixin):
    xs: frozenset


@dataclass(frozen=True)
class DataClassWithFrozenSetUnbound(DataClassJsonMixin):
    xs: FrozenSet


@dataclass(frozen=True)
class DataClassWithDeque(DataClassJsonMixin):
    xs: Deque[int]


@dataclass(frozen=True)
class DataClassWithDequeCollections(DataClassJsonMixin):
    xs: deque


@dataclass(frozen=True)
class DataClassWithDequeUnbound(DataClassJsonMixin):
    xs: Deque[int]


@dataclass(frozen=True)
class DataClassWithOptional(DataClassJsonMixin):
    x: Optional[int]


@dataclass(frozen=True)
class DataClassWithOptionalWithDefault(DataClassJsonMixin):
    x: Optional[int] = None


@dataclass(frozen=True)
class DataClassWithOptionalUnbound(DataClassJsonMixin):
    x: Optional


@dataclass
class DataClassWithOptionalStr(DataClassJsonMixin):
    x: Optional[str] = None


@dataclass(frozen=True)
class DataClassWithOptionalNested(DataClassJsonMixin):
    x: Optional[DataClassWithOptional]


@dataclass(frozen=True)
class DataClassWithOptionalNestedWithDefault(DataClassJsonMixin):
    x: Optional[DataClassWithOptional] = None


@dataclass(frozen=True)
class DataClassWithUnionIntNone(DataClassJsonMixin):
    x: Union[int, None]


@dataclass(frozen=True)
class DataClassWithDataClass(DataClassJsonMixin):
    dc_with_list: DataClassWithList


@dataclass(frozen=True)
class DataClassX(DataClassJsonMixin):
    x: int


@dataclass(frozen=True)
class DataClassXs(DataClassJsonMixin):
    xs: List[DataClassX]


@dataclass(frozen=True)
class DataClassIntImmutableDefault(DataClassJsonMixin):
    x: int = 0


@dataclass(frozen=True)
class DataClassBoolImmutableDefault(DataClassJsonMixin):
    x: bool = False


@dataclass(frozen=True)
class DataClassMutableDefaultList(DataClassJsonMixin):
    xs: List[int] = field(default_factory=list)


@dataclass(frozen=True)
class DataClassMutableDefaultDict(DataClassJsonMixin):
    xs: Dict[str, int] = field(default_factory=dict)


class MyCollection(Collection[A]):
    def __init__(self, xs: Collection[A]) -> None:
        self.xs = xs

    def __contains__(self, item):
        return False

    def __iter__(self):
        return iter(self.xs)

    def __len__(self):
        return len(self.xs)

    def __eq__(self, other):
        return type(self) == type(other) and self.xs == other.xs


@dataclass(frozen=True)
class DataClassWithMyCollection(DataClassJsonMixin):
    xs: MyCollection[int]


@dataclass_json
@dataclass
class DataClassJsonDecorator:
    x: str


@dataclass_json
@dataclass
class DataClassWithConfigManual:
    id: float = field(
        metadata={'dataclasses_json': {
            'mm_field': fields.Integer()
        }})


@dataclass_json
@dataclass
class DataClassWithConfigHelper:
    id: float = field(metadata=config(encoder=str))


@dataclass_json
@dataclass
class DataClassWithErroneousDecode:
    # Accepts no arguments, so passing in a single argument will result in a TypeError.
    id: float = field(metadata=config(decoder=lambda: None))


def split_str(data: str, *_args, **_kwargs):
    return data.split(',')


@dataclass_json
@dataclass
class DataClassDifferentTypeDecode:
    lst: List[str] = field(default=None, metadata=config(decoder=split_str))


@dataclass_json
@dataclass
class DataClassMappingBadDecode:
    map: Dict[str, DataClassWithErroneousDecode]


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class DataClassWithConfigDecorator:
    id_field: str


@dataclass_json
@dataclass
class DataClassWithUuid:
    id: UUID


@dataclass_json
@dataclass
class DataClassDefaultListStr:
    value: List[str] = field(default_factory=list)


@dataclass_json
@dataclass
class DataClassChild:
    name: str


@dataclass_json
@dataclass
class DataClassDefaultOptionalList:
    children: Optional[List[DataClassChild]] = None


@dataclass_json
@dataclass
class DataClassList:
    children: List[DataClassChild]


@dataclass_json
@dataclass
class DataClassOptional:
    a: int
    b: Optional[int]


@dataclass_json
@dataclass
class DataClassWithOptionalDatetime:
    a: Optional[datetime]


@dataclass_json
@dataclass
class DataClassWithOptionalDecimal:
    a: Optional[Decimal]


@dataclass_json
@dataclass
class DataClassWithOptionalUuid:
    a: Optional[UUID]


@dataclass_json
@dataclass
class DataClassWithNestedAny:
    a: Dict[str, Any]


@dataclass_json
@dataclass
class DataClassWithNestedOptionalAny:
    a: Dict[str, Optional[Any]]


@dataclass_json
@dataclass
class DataClassWithNestedOptional:
    a: Dict[str, Optional[int]]


@dataclass_json
@dataclass
class DataClassWithNestedDictWithTupleKeys:
    a: Dict[Tuple[int], int]


@dataclass_json
@dataclass
class DataClassWithCounter:
    c: Counter[str]


@dataclass_json
@dataclass
class DataClassWithSelf(DataClassJsonMixin):
    id: str
    ref: Optional['DataClassWithSelf']



from collections import deque

from hypothesis import given
from hypothesis.strategies import (frozensets, integers, lists, one_of, sets,
                                   tuples)

from tests.entities import (DataClassWithDeque, DataClassWithFrozenSet,
                            DataClassWithList, DataClassWithOptional,
                            DataClassWithSet, DataClassWithTuple)
from tests.hypothesis2 import examples
from tests.hypothesis2.strategies import deques, optionals

dcconss_strategies_conss = [(DataClassWithList, lists, list),
                            (DataClassWithSet, sets, set),
                            (DataClassWithTuple, tuples, tuple),
                            (DataClassWithFrozenSet, frozensets, frozenset),
                            (DataClassWithDeque, deques, deque),
                            (DataClassWithOptional, optionals, lambda x: x[0])]
example_input = [1]


@given(one_of(*[strategy_fn(integers()).map(dccons)
                for dccons, strategy_fn, _ in dcconss_strategies_conss]))
@examples(*[dccons(cons(example_input))
            for dccons, _, cons in dcconss_strategies_conss])
def test_generic_encode_and_decode_are_inverses(dc):
    assert dc.from_json(dc.to_json()) == dc



from dataclasses import dataclass, field

import pytest

from dataclasses_json import LetterCase, dataclass_json, config


@dataclass_json
@dataclass
class CamelCasePerson:
    given_name: str = field(
        metadata={'dataclasses_json': {
            'letter_case': LetterCase.CAMEL
        }}
    )


@dataclass_json
@dataclass
class KebabCasePerson:
    given_name: str = field(
        metadata={'dataclasses_json': {
            'letter_case': LetterCase.KEBAB
        }}
    )


@dataclass_json
@dataclass
class SnakeCasePerson:
    given_name: str = field(
        metadata={'dataclasses_json': {
            'letter_case': LetterCase.SNAKE
        }}
    )


@dataclass_json
@dataclass
class PascalCasePerson:
    given_name: str = field(
        metadata={'dataclasses_json': {
            'letter_case': LetterCase.PASCAL
        }}
    )


@dataclass_json
@dataclass
class FieldNamePerson:
    given_name: str = field(metadata=config(field_name='givenName'))


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class CamelCasePersonWithOverride:
    given_name: str
    years_on_earth: int = field(metadata=config(field_name='age'))


@dataclass_json
@dataclass
class CamelCaseProtectedNamePerson:
    _given_name_2: str = field(
        metadata={'dataclasses_json': {
            'letter_case': LetterCase.CAMEL
        }}
    )


@dataclass_json
@dataclass
class CamelCaseDuplicatedNameEncodingPerson:
    given_name_1: str = field(
        metadata={'dataclasses_json': {
            'letter_case': LetterCase.CAMEL
        }}
    )
    given_name1: str = field(
        metadata={'dataclasses_json': {
            'letter_case': LetterCase.CAMEL
        }}
    )


class TestLetterCase:
    def test_camel_encode(self):
        assert CamelCasePerson('Alice').to_json() == '{"givenName": "Alice"}'

    def test_camel_decode(self):
        assert CamelCasePerson.from_json(
            '{"givenName": "Alice"}') == CamelCasePerson('Alice')

    def test_kebab_encode(self):
        assert KebabCasePerson('Alice').to_json() == '{"given-name": "Alice"}'

    def test_kebab_decode(self):
        assert KebabCasePerson.from_json(
            '{"given-name": "Alice"}') == KebabCasePerson('Alice')

    def test_snake_encode(self):
        assert SnakeCasePerson('Alice').to_json() == '{"given_name": "Alice"}'

    def test_snake_decode(self):
        assert SnakeCasePerson.from_json(
            '{"given_name": "Alice"}') == SnakeCasePerson('Alice')

    def test_pascal_encode(self):
        assert PascalCasePerson('Alice').to_json() == '{"GivenName": "Alice"}'

    def test_pascal_decode(self):
        assert PascalCasePerson.from_json(
            '{"GivenName": "Alice"}') == PascalCasePerson('Alice')

    def test_field_name_encode(self):
        assert FieldNamePerson('Alice').to_json() == '{"givenName": "Alice"}'

    def test_field_name_decode(self):
        assert FieldNamePerson.from_json(
            '{"givenName": "Alice"}') == FieldNamePerson('Alice')

    def test_camel_with_override_encode(self):
        assert CamelCasePersonWithOverride(
            'Alice', 10).to_json() == '{"givenName": "Alice", "age": 10}'

    def test_camel_with_override_decode(self):
        expected = CamelCasePersonWithOverride('Alice', 10)
        assert CamelCasePersonWithOverride.from_json(
            '{"givenName": "Alice", "age": 10}') == expected

    def test_from_dict(self):
        assert CamelCasePerson.from_dict(
            {'givenName': 'Alice'}) == CamelCasePerson('Alice')

    def test_to_dict(self):
        assert {'givenName': 'Alice'} == CamelCasePerson('Alice').to_dict()

    def test_protected_encode(self):
        assert CamelCaseProtectedNamePerson('Alice').to_json() == '{"givenName2": "Alice"}'

    def test_protected_decode(self):
        assert CamelCaseProtectedNamePerson.from_json('{"givenName2": "Alice"}') == CamelCaseProtectedNamePerson('Alice')

    def test_duplicated_encoding(self):
        with pytest.raises(ValueError):
            CamelCaseDuplicatedNameEncodingPerson('Alice', 'Bob').to_json()



from dataclasses import dataclass

from dataclasses_json import DataClassJsonMixin


class MyStr(str):

    def is_even_length(self) -> bool:
        return len(self) % 2 == 0


@dataclass(frozen=True)
class DataClassWithStrSubclass(DataClassJsonMixin):
    any_str: str
    my_str: MyStr


class TestDataClassWithStrSubclass:

    def test_encode__no_instantiation_required(self):
        model_dict = {"any_str": "str", "my_str": MyStr("str")}
        expected = DataClassWithStrSubclass(any_str="str", my_str=MyStr("str"))
        actual = DataClassWithStrSubclass.from_dict(model_dict)
        assert expected == actual
        assert model_dict["my_str"] is actual.my_str

    def test_encode__subclass_str_instantiated(self):
        model_dict = {"any_str": "str", "my_str": "str"}
        expected = DataClassWithStrSubclass(any_str="str", my_str=MyStr("str"))
        actual = DataClassWithStrSubclass.from_dict(model_dict)
        assert expected == actual
        assert model_dict["my_str"] is not actual.my_str


from dataclasses import dataclass, field

from dataclasses_json.api import DataClassJsonMixin, config
from dataclasses_json.cfg import Exclude


@dataclass
class EncodeExclude(DataClassJsonMixin):
    public_field: str
    private_field: str = field(metadata=config(exclude=Exclude.ALWAYS))


@dataclass
class EncodeInclude(DataClassJsonMixin):
    public_field: str
    private_field: str = field(metadata=config(exclude=Exclude.NEVER))


@dataclass
class EncodeCustom(DataClassJsonMixin):
    public_field: str
    sensitive_field: str = field(
        metadata=config(exclude=lambda v: v.startswith("secret"))
    )


def test_exclude():
    dclass = EncodeExclude(public_field="public", private_field="private")
    encoded = dclass.to_dict()
    assert "public_field" in encoded
    assert "private_field" not in encoded


def test_include():
    dclass = EncodeInclude(public_field="public", private_field="private")
    encoded = dclass.to_dict()
    assert "public_field" in encoded
    assert "private_field" in encoded
    assert encoded["private_field"] == "private"


def test_custom_action_included():
    dclass = EncodeCustom(public_field="public", sensitive_field="notsecret")
    encoded = dclass.to_dict()
    assert "sensitive_field" in encoded


def test_custom_action_excluded():
    dclass = EncodeCustom(public_field="public", sensitive_field="secret")
    encoded = dclass.to_dict()
    assert "sensitive_field" not in encoded



import datetime
from dataclasses import dataclass, field
from typing import Optional

import pytest
from marshmallow import fields, ValidationError

from dataclasses_json import DataClassJsonMixin


@dataclass
class Car(DataClassJsonMixin):
    license_number: str = field(
        metadata={'dataclasses_json': {
            'mm_field': fields.String(required=False)}
        })


@dataclass
class StringDate(DataClassJsonMixin):
    string_date: datetime.datetime = field(
        metadata={'dataclasses_json': {
            'encoder': str,
            'decoder': str,
            'mm_field': fields.String(required=False)}
        })
    
@dataclass
class OptionalStringDate(DataClassJsonMixin):
    string_date: Optional[datetime.datetime] = field(
        default=None,
        metadata={'dataclasses_json': {
            'encoder': str,
            'decoder': str,
            'mm_field': fields.String(required=False)}
        })


car_schema = Car.schema()
string_date_schema = StringDate.schema()
opt_string_date_schema = OptionalStringDate.schema()


class TestMetadata:
    def test_validation_error_raises(self):
        with pytest.raises(ValidationError) as e:
            car_schema.load({'license_number': 123})
        assert e.value.messages == {'license_number': ['Not a valid string.']}

    def test_mm_field_takes_precedence_over_types(self):
        obj = string_date_schema.load({'string_date': 'yesterday'})
        assert isinstance(obj, StringDate)
        assert obj.string_date == 'yesterday'

    def test_optional_field_only_decoded_when_present(self):
        obj = opt_string_date_schema.load({})
        assert isinstance(obj, OptionalStringDate)
        assert obj.string_date == None
        
        another_obj = opt_string_date_schema.load({'string_date': 'today'})
        assert isinstance(another_obj, OptionalStringDate)
        assert another_obj.string_date == 'today'



import sys
from datetime import datetime, timezone
from decimal import Decimal
from uuid import UUID

import pytest

from tests.entities import (DataClassBoolImmutableDefault,
                            DataClassIntImmutableDefault,
                            DataClassJsonDecorator,
                            DataClassWithConfigDecorator,
                            DataClassWithConfigHelper,
                            DataClassWithConfigManual, DataClassWithDataClass,
                            DataClassWithDecimal, DataClassWithList,
                            DataClassWithNestedNewType, DataClassWithNewType,
                            DataClassWithOptional,
                            DataClassWithOptionalDatetime,
                            DataClassWithOptionalDecimal,
                            DataClassWithOptionalNested,
                            DataClassWithOptionalUuid, DataClassWithUuid, UUIDWrapper,
                            UUIDWrapperWrapper,
                            DataClassWithOptionalWithDefault,
                            DataClassWithOptionalNestedWithDefault)


class TestTypes:
    decimal_s = "12345.12345"
    dc_decimal_json = f'{{"x": "{decimal_s}"}}'

    uuid_s = 'd1d61dd7-c036-47d3-a6ed-91cc2e885fc8'
    dc_uuid_json = f'{{"id": "{uuid_s}"}}'

    def test_decimal_encode(self):
        assert (DataClassWithDecimal(Decimal(self.decimal_s)).to_json()
                == self.dc_decimal_json)

    def test_decimal_decode(self):
        assert (DataClassWithDecimal.from_json(self.dc_decimal_json)
                == DataClassWithDecimal(Decimal(self.decimal_s)))


class TestGenericExtendedTypes:
    def test_optional_datetime(self):
        dt = datetime(2018, 11, 17, 16, 55, 28, 456753, tzinfo=timezone.utc)
        dc = DataClassWithOptionalDatetime(dt)
        assert (DataClassWithOptionalDatetime.from_json(dc.to_json())
                == dc)

    def test_optional_decimal(self):
        dc = DataClassWithOptionalDecimal(Decimal("12345.12345"))
        assert (DataClassWithOptionalDecimal.from_json(dc.to_json())
                == dc)

    def test_optional_uuid(self):
        dc = DataClassWithOptionalUuid(
            UUID('d1d61dd7-c036-47d3-a6ed-91cc2e885fc8'))
        assert (DataClassWithOptionalUuid.from_json(dc.to_json())
                == dc)


class TestDictDecode:
    decimal_s = "12345.12345"
    dc_decimal_json = {"x": decimal_s}

    uuid_s = 'd1d61dd7-c036-47d3-a6ed-91cc2e885fc8'
    dc_uuid_json = {"id": uuid_s}

    def test_decimal_decode(self):
        assert (DataClassWithDecimal.from_dict(self.dc_decimal_json)
                == DataClassWithDecimal(Decimal(self.decimal_s)))

    def test_uuid_decode(self):
        assert (DataClassWithUuid.from_dict(self.dc_uuid_json)
                == DataClassWithUuid(UUID(self.uuid_s)))


class TestNewType:
    new_type_s = 'd1d61dd7-c036-47d3-a6ed-91cc2e885fc8'
    dc_new_type_json = f'{{"id": "{new_type_s}"}}'

    def test_new_type_encode(self):
        assert (DataClassWithNewType(UUIDWrapper(UUID(self.new_type_s))).to_json()
                == self.dc_new_type_json)

    @pytest.mark.skipif(sys.version_info >= (3, 10),
                        reason="newtype decode breaks in 3.10")
    def test_new_type_decode(self):
        assert (DataClassWithNewType.from_json(self.dc_new_type_json)
                == DataClassWithNewType(UUIDWrapper(UUID(self.new_type_s))))

    def test_nested_new_type_encode(self):
        assert (DataClassWithNestedNewType(
            UUIDWrapperWrapper(UUIDWrapper(UUID(self.new_type_s)))).to_json()
                == self.dc_new_type_json)

    @pytest.mark.skipif(sys.version_info >= (3, 10),
                        reason="newtype decode breaks in 3.10")
    def test_nested_new_type_decode(self):
        assert (DataClassWithNestedNewType.from_json(self.dc_new_type_json)
                == DataClassWithNestedNewType(
                    UUIDWrapperWrapper(UUIDWrapper(UUID(self.new_type_s)))))


class TestInferMissing:
    def test_infer_missing(self):
        actual = DataClassWithOptional.from_json('{}', infer_missing=True)
        assert (actual == DataClassWithOptional(None))

    def test_infer_missing_is_recursive(self):
        actual = DataClassWithOptionalNested.from_json('{"x": {}}',
                                                       infer_missing=True)
        expected = DataClassWithOptionalNested(DataClassWithOptional(None))
        assert (actual == expected)

    def test_infer_missing_terminates_at_first_missing(self):
        actual = DataClassWithOptionalNested.from_json('{"x": null}',
                                                       infer_missing=True)
        assert (actual == DataClassWithOptionalNested(None))


class TestWarnings:
    def test_warns_when_nonoptional_field_is_missing_with_infer_missing(self):
        with pytest.warns(RuntimeWarning, match='Missing value'):
            actual = DataClassWithDataClass.from_json('{"dc_with_list": {}}',
                                                      infer_missing=True)
            expected = DataClassWithDataClass(DataClassWithList(None))
            assert (actual == expected)

    def test_warns_when_required_field_is_none(self):
        with pytest.warns(RuntimeWarning, match='\'NoneType\' object'):
            assert (DataClassWithDataClass.from_json(
                '{"dc_with_list": null}') == DataClassWithDataClass(None))


class TestErrors:
    def test_error_when_nonoptional_field_is_missing(self):
        with pytest.raises(KeyError):
            actual = DataClassWithDataClass.from_json('{"dc_with_list": {}}')
            expected = DataClassWithDataClass(DataClassWithList(None))
            assert (actual == expected)


class TestDecorator:
    def test_decorator(self):
        json_s = '{"x": "x"}'
        assert DataClassJsonDecorator.from_json(json_s).to_json() == json_s


class TestSchema:
    def test_dumps_many(self):
        actual = DataClassWithList.schema().dumps([DataClassWithList([1])],
                                                  many=True)
        json_s = '[{"xs": [1]}]'
        assert actual == json_s

    def test_dumps_many_nested(self):
        dumped = DataClassWithDataClass.schema().dumps(
            [DataClassWithDataClass(DataClassWithList([1]))],
            many=True)
        json_s = '[{"dc_with_list": {"xs": [1]}}]'
        assert dumped == json_s

    def test_loads_many(self):
        json_s = '[{"xs": [1]}]'
        assert (DataClassWithList.schema().loads(json_s, many=True)
                == [DataClassWithList([1])])

    def test_loads_many_nested(self):
        json_s = '[{"dc_with_list": {"xs": [1]}}]'
        assert (DataClassWithDataClass.schema().loads(json_s, many=True)
                == [DataClassWithDataClass(DataClassWithList([1]))])

    def test_loads_default(self):
        assert (DataClassIntImmutableDefault.schema().loads('{}')
                == DataClassIntImmutableDefault())
        assert (DataClassBoolImmutableDefault.schema().loads('{}')
                == DataClassBoolImmutableDefault())

    def test_loads_default_many(self):
        assert (DataClassIntImmutableDefault.schema().loads('[{}]', many=True)
                == [DataClassIntImmutableDefault()])
        assert (DataClassBoolImmutableDefault.schema().loads('[{}]', many=True)
                == [DataClassBoolImmutableDefault()])

    def test_dumps_default(self):
        d_int = DataClassIntImmutableDefault()
        assert d_int.x == 0
        assert DataClassIntImmutableDefault.schema().dumps(d_int) == '{"x": 0}'
        d_bool = DataClassBoolImmutableDefault()
        assert d_bool.x is False
        assert DataClassBoolImmutableDefault.schema().dumps(
            d_bool) == '{"x": false}'

    def test_dumps_default_many(self):
        d_int = DataClassIntImmutableDefault()
        assert d_int.x == 0
        assert DataClassIntImmutableDefault.schema().dumps([d_int],
                                                           many=True) == '[{"x": 0}]'
        d_bool = DataClassBoolImmutableDefault()
        assert d_bool.x is False
        assert DataClassBoolImmutableDefault.schema().dumps([d_bool],
                                                            many=True) == '[{"x": false}]'

    def test_dumps_new_type(self):
        raw_value = 'd1d61dd7-c036-47d3-a6ed-91cc2e885fc8'
        id_value = UUIDWrapper(UUID(raw_value))

        d_new_type = DataClassWithNewType(id_value)

        assert DataClassWithNewType.schema().dumps(
            d_new_type) == f'{{"id": "{raw_value}"}}'

    def test_dumps_nested_new_type(self):
        raw_value = 'd1d61dd7-c036-47d3-a6ed-91cc2e885fc8'
        id_value = UUIDWrapperWrapper(UUIDWrapper(UUID(raw_value)))

        d_new_type = DataClassWithNestedNewType(id_value)

        assert DataClassWithNestedNewType.schema().dumps(
            d_new_type) == f'{{"id": "{raw_value}"}}'

    def test_loads_infer_missing(self):
        assert (DataClassWithOptionalWithDefault
                .schema(infer_missing=True)
                .loads('[{}]', many=True) == [DataClassWithOptionalWithDefault(None)])

    def test_loads_infer_missing_nested(self):
        assert (DataClassWithOptionalNestedWithDefault
                .schema(infer_missing=True)
                .loads('[{}]', many=True) == [
                    DataClassWithOptionalNestedWithDefault(None)])


class TestOverride:
    def test_override(self):
        dc = DataClassWithConfigManual(5.0)
        assert dc.to_json() == '{"id": 5.0}'
        assert DataClassWithConfigManual.schema().dumps(dc) == '{"id": 5}'

    def test_override_with_config_helper(self):
        dc = DataClassWithConfigHelper(5.0)
        assert dc.to_json() == '{"id": "5.0"}'


class TestConfig:
    def test_config_encode(self):
        dc = DataClassWithConfigDecorator('a')
        assert dc.to_json() == '{"idField": "a"}'

    def test_config_decode(self):
        dc = DataClassWithConfigDecorator('a')
        assert DataClassWithConfigDecorator.from_json('{"idField": "a"}') == dc



from collections import Counter, deque

from tests.entities import (DataClassIntImmutableDefault,
                            DataClassMutableDefaultDict,
                            DataClassMutableDefaultList, DataClassWithDeque,
                            DataClassWithDict, DataClassWithDictInt,
                            DataClassWithDictUnbound,
                            DataClassWithFrozenSet, DataClassWithList,
                            DataClassWithListUnbound, DataClassWithListStr, DataClassWithMyCollection,
                            DataClassWithOptional, DataClassWithOptionalStr,
                            DataClassWithSet, DataClassWithSetUnbound,
                            DataClassWithOptionalUnbound,
                            DataClassWithSetBuiltin,
                            DataClassWithDictBuiltin,
                            DataClassWithListBuiltin,
                            DataClassWithTupleBuiltin,
                            DataClassWithFrozenSetBuiltin,
                            DataClassWithDequeUnbound,
                            DataClassWithFrozenSetUnbound,
                            DataClassWithDequeCollections,
                            DataClassWithTuple, DataClassWithTupleUnbound,
                            DataClassWithUnionIntNone, MyCollection,
                            DataClassWithCounter)


class TestEncoder:
    def test_list(self):
        assert DataClassWithList([1]).to_json() == '{"xs": [1]}'

    def test_list_unbound(self):
        assert DataClassWithListUnbound([1]).to_json() == '{"xs": [1]}'

    def test_list_builtin(self):
        assert DataClassWithListBuiltin([1]).to_json() == '{"xs": [1]}'

    def test_list_str(self):
        assert DataClassWithListStr(['1']).to_json() == '{"xs": ["1"]}'

    def test_dict(self):
        assert DataClassWithDict({'1': 'a'}).to_json() == '{"kvs": {"1": "a"}}'

    def test_dict_unbound(self):
        assert DataClassWithDictUnbound({'1': 'a'}).to_json() == '{"kvs": {"1": "a"}}'

    def test_dict_builtin(self):
        assert DataClassWithDictBuiltin({'1': 'a'}).to_json() == '{"kvs": {"1": "a"}}'

    def test_dict_int(self):
        assert DataClassWithDictInt({1: 'a'}).to_json() == '{"kvs": {"1": "a"}}'

    def test_set(self):
        assert DataClassWithSet({1}).to_json() == '{"xs": [1]}'

    def test_set_unbound(self):
        assert DataClassWithSetUnbound({1}).to_json() == '{"xs": [1]}'

    def test_set_builtin(self):
        assert DataClassWithSetBuiltin({1}).to_json() == '{"xs": [1]}'

    def test_tuple(self):
        assert DataClassWithTuple((1,)).to_json() == '{"xs": [1]}'

    def test_tuple_unbound(self):
        assert DataClassWithTupleUnbound((1,)).to_json() == '{"xs": [1]}'

    def test_tuple_builtin(self):
        assert DataClassWithTupleBuiltin((1,)).to_json() == '{"xs": [1]}'

    def test_frozenset(self):
        assert DataClassWithFrozenSet(frozenset([1])).to_json() == '{"xs": [1]}'

    def test_frozenset_unbound(self):
        assert DataClassWithFrozenSetUnbound(frozenset([1])).to_json() == '{"xs": [1]}'

    def test_frozenset_builtin(self):
        assert DataClassWithFrozenSetBuiltin(frozenset([1])).to_json() == '{"xs": [1]}'

    def test_deque(self):
        assert DataClassWithDeque(deque([1])).to_json() == '{"xs": [1]}'

    def test_deque_unbound(self):
        assert DataClassWithDequeUnbound(deque([1])).to_json() == '{"xs": [1]}'

    def test_deque_builtin(self):
        assert DataClassWithDequeCollections(deque([1])).to_json() == '{"xs": [1]}'

    def test_optional(self):
        assert DataClassWithOptional(1).to_json() == '{"x": 1}'
        assert DataClassWithOptional(None).to_json() == '{"x": null}'

    def test_optional_unbound(self):
        assert DataClassWithOptionalUnbound(1).to_json() == '{"x": 1}'

    def test_optional_str(self):
        assert DataClassWithOptionalStr('1').to_json() == '{"x": "1"}'
        assert DataClassWithOptionalStr(None).to_json() == '{"x": null}'
        assert DataClassWithOptionalStr().to_json() == '{"x": null}'

    def test_union_int_none(self):
        assert DataClassWithUnionIntNone(1).to_json() == '{"x": 1}'
        assert DataClassWithUnionIntNone(None).to_json() == '{"x": null}'

    def test_my_collection(self):
        assert DataClassWithMyCollection(
            MyCollection([1])).to_json() == '{"xs": [1]}'

    def test_immutable_default(self):
        assert DataClassIntImmutableDefault().to_json() == '{"x": 0}'

    def test_mutable_default_list(self):
        assert DataClassMutableDefaultList().to_json() == '{"xs": []}'

    def test_mutable_default_dict(self):
        assert DataClassMutableDefaultDict().to_json() == '{"xs": {}}'

    def test_counter(self):
        assert DataClassWithCounter(
            c=Counter('foo')).to_json() == '{"c": {"f": 1, "o": 2}}'


class TestDecoder:
    def test_list(self):
        assert (DataClassWithList.from_json('{"xs": [1]}') ==
                DataClassWithList([1]))

    def test_list_unbound(self):
        assert (DataClassWithListUnbound.from_json('{"xs": [1]}') ==
                DataClassWithListUnbound([1]))

    def test_list_builtin(self):
        assert (DataClassWithListBuiltin.from_json('{"xs": [1]}') ==
                DataClassWithListBuiltin([1]))

    def test_list_str(self):
        assert (DataClassWithListStr.from_json('{"xs": ["1"]}') ==
                DataClassWithListStr(["1"]))

    def test_dict(self):
        assert (DataClassWithDict.from_json('{"kvs": {"1": "a"}}') ==
                DataClassWithDict({'1': 'a'}))

    def test_dict_unbound(self):
        assert (DataClassWithDictUnbound.from_json('{"kvs": {"1": "a"}}') ==
                DataClassWithDictUnbound({'1': 'a'}))

    def test_dict_builtin(self):
        assert (DataClassWithDictBuiltin.from_json('{"kvs": {"1": "a"}}') ==
                DataClassWithDictBuiltin({'1': 'a'}))

    def test_dict_int(self):
        assert (DataClassWithDictInt.from_json('{"kvs": {"1": "a"}}') ==
                DataClassWithDictInt({1: 'a'}))

    def test_set(self):
        assert (DataClassWithSet.from_json('{"xs": [1]}') ==
                DataClassWithSet({1}))

    def test_set_unbound(self):
        assert (DataClassWithSetUnbound.from_json('{"xs": [1]}') ==
                DataClassWithSetUnbound({1}))

    def test_set_builtin(self):
        assert (DataClassWithSetBuiltin.from_json('{"xs": [1]}') ==
                DataClassWithSetBuiltin({1}))

    def test_tuple(self):
        assert (DataClassWithTuple.from_json('{"xs": [1]}') ==
                DataClassWithTuple((1,)))

    def test_tuple_unbound(self):
        assert (DataClassWithTupleUnbound.from_json('{"xs": [1]}') ==
                DataClassWithTupleUnbound((1,)))

    def test_tuple_builtin(self):
        assert (DataClassWithTupleBuiltin.from_json('{"xs": [1]}') ==
                DataClassWithTupleBuiltin((1,)))

    def test_frozenset(self):
        assert (DataClassWithFrozenSet.from_json('{"xs": [1]}') ==
                DataClassWithFrozenSet(frozenset([1])))

    def test_frozenset_unbound(self):
        assert (DataClassWithFrozenSetUnbound.from_json('{"xs": [1]}') ==
                DataClassWithFrozenSetUnbound(frozenset([1])))

    def test_frozenset_builtin(self):
        assert (DataClassWithFrozenSetBuiltin.from_json('{"xs": [1]}') ==
                DataClassWithFrozenSetBuiltin(frozenset([1])))

    def test_deque(self):
        assert (DataClassWithDeque.from_json('{"xs": [1]}') ==
                DataClassWithDeque(deque([1])))

    def test_deque_unbound(self):
        assert (DataClassWithDequeUnbound.from_json('{"xs": [1]}') ==
                DataClassWithDequeUnbound(deque([1])))

    def test_deque_collections(self):
        assert (DataClassWithDequeCollections.from_json('{"xs": [1]}') ==
                DataClassWithDequeCollections(deque([1])))

    def test_optional(self):
        assert (DataClassWithOptional.from_json('{"x": 1}') ==
                DataClassWithOptional(1))
        assert (DataClassWithOptional.from_json('{"x": null}') ==
                DataClassWithOptional(None))

    def test_optional_unbound(self):
        assert (DataClassWithOptionalUnbound.from_json('{"x": 1}') ==
                DataClassWithOptionalUnbound(1))
        assert (DataClassWithOptionalUnbound.from_json('{"x": null}') ==
                DataClassWithOptionalUnbound(None))

    def test_optional_str(self):
        assert (DataClassWithOptionalStr.from_json('{"x": "1"}') ==
                DataClassWithOptionalStr("1"))
        assert (DataClassWithOptionalStr.from_json('{"x": null}') ==
                DataClassWithOptionalStr(None))
        assert (DataClassWithOptionalStr.from_json('{}', infer_missing=True) ==
                DataClassWithOptionalStr())

    def test_my_collection(self):
        assert (DataClassWithMyCollection.from_json('{"xs": [1]}') ==
                DataClassWithMyCollection(MyCollection([1])))

    def test_immutable_default(self):
        assert (DataClassIntImmutableDefault.from_json('{"x": 0}')
                == DataClassIntImmutableDefault())
        assert (DataClassMutableDefaultList.from_json('{}', infer_missing=True)
                == DataClassMutableDefaultList())

    def test_mutable_default_list(self):
        assert (DataClassMutableDefaultList.from_json('{"xs": []}')
                == DataClassMutableDefaultList())
        assert (DataClassMutableDefaultList.from_json('{}', infer_missing=True)
                == DataClassMutableDefaultList())

    def test_mutable_default_dict(self):
        assert (DataClassMutableDefaultDict.from_json('{"kvs": {}}')
                == DataClassMutableDefaultDict())
        assert (DataClassMutableDefaultDict.from_json('{}', infer_missing=True)
                == DataClassMutableDefaultDict())

    def test_counter(self):
        assert DataClassWithCounter.from_json('{"c": {"f": 1, "o": 2}}') == \
               DataClassWithCounter(c=Counter('foo'))



from dataclasses import dataclass
from decimal import Decimal
from typing import Optional

from pytest import mark, param


from dataclasses_json import DataClassJsonMixin


@dataclass(frozen=True)
class DataClassWithBuiltins(DataClassJsonMixin):
    actually_a_str: str
    actually_an_int: int
    actually_a_float: float


@mark.parametrize(
    "model_dict, expected_model",
    [
        param(
            {"actually_a_str": "str", "actually_an_int": 42, "actually_a_float": 42.1},
            DataClassWithBuiltins(actually_a_str="str", actually_an_int=42, actually_a_float=42.1),
            id="Happy case"
        ),
        param(
            {"actually_a_str": "str", "actually_an_int": Decimal("42.1"), "actually_a_float": Decimal("42.1")},
            DataClassWithBuiltins(actually_a_str="str", actually_an_int=42, actually_a_float=42.1),
            id="Decimal as int and float"
        ),
    ]
)
def test__DataClassWithBuiltins__from_dict(model_dict, expected_model):
    assert DataClassWithBuiltins.from_dict(model_dict) == expected_model



from typing import Dict, Union, Any
import pytest

from dataclasses import dataclass

from dataclasses_json import dataclass_json


@dataclass_json
@dataclass(frozen=True)
class DataWithDict:
    metadata: Dict[Any, Any]


@dataclass_json
@dataclass(frozen=True)
class DataWithTypedDict:
    metadata: Dict[str, Union[str, float]]


@dataclass_json
@dataclass(frozen=True)
class DataWithPythonDict:
    metadata: dict


example_metadata_dict = {"some_data": "written_here", "some_score": 34.4}
d_ex_1 = DataWithDict(metadata=example_metadata_dict)
d_ex_2 = DataWithTypedDict(metadata=example_metadata_dict)
d_ex_3 = DataWithPythonDict(metadata=example_metadata_dict)
d_ex_as_dict = {"metadata": example_metadata_dict}


class TestEncoder:
    def test_dataclass_with_dict(self):
        assert d_ex_1.to_dict() == d_ex_as_dict, f'Actual: {d_ex_1.to_dict()}, Expected: {d_ex_as_dict}'

    def test_dataclass_with_typed_dict(self):
        assert d_ex_2.to_dict() == d_ex_as_dict, f'Actual: {d_ex_2.to_dict()}, Expected: {d_ex_as_dict}'

    def test_dataclass_with_python_dict(self):
        assert d_ex_3.to_dict() == d_ex_as_dict, f'Actual: {d_ex_3.to_dict()}, Expected: {d_ex_as_dict}'


class TestDecoder:
    def test_dataclass_with_dict(self):
        d_from_dict = DataWithDict.from_dict(d_ex_as_dict)
        assert d_ex_1 == d_from_dict, f'Actual: {d_from_dict}, Expected: {d_ex_1}'

    def test_dataclass_with_typed_dict(self):
        d_from_dict = DataWithTypedDict.from_dict(d_ex_as_dict)
        assert d_ex_2 == d_from_dict, f'Actual: {d_from_dict}, Expected: {d_ex_2}'

    def test_dataclass_with_python_dict(self):
        d_from_dict = DataWithPythonDict.from_dict(d_ex_as_dict)
        assert d_ex_3 == d_from_dict, f'Actual: {d_from_dict}, Expected: {d_ex_3}'


class TestValidator:
    @pytest.mark.parametrize('metadata_dict, is_valid', [
        ({}, True),
        (example_metadata_dict, True),
        ({"is_boolean": True}, True),
        (None, False)
    ])
    def test_dataclass_with_dict(self, metadata_dict, is_valid):
        schema = DataWithDict.schema()
        res = schema.validate({"metadata": metadata_dict})
        assert not res == is_valid

    @pytest.mark.parametrize('metadata_dict, is_valid', [
        ({}, True),
        (example_metadata_dict, True),
        ({"is_boolean": True}, False),
        ({5: "a number"}, False)
    ])
    def test_dataclass_with_typed_dict(self, metadata_dict, is_valid):
        schema = DataWithTypedDict.schema()
        res = schema.validate({"metadata": metadata_dict})
        assert not res == is_valid

    @pytest.mark.parametrize('metadata_dict, is_valid', [
        ({}, True),
        (example_metadata_dict, True),
        ({"is_boolean": True}, True),
        (None, False)
    ])
    def test_dataclass_with_python_dict(self, metadata_dict, is_valid):
        schema = DataWithPythonDict.schema()
        res = schema.validate({"metadata": metadata_dict})
        assert not res == is_valid



from dataclasses import dataclass
from typing import Set, Optional

from dataclasses_json import dataclass_json


@dataclass_json
@dataclass(frozen=True)
class Student:
    id: int = 0
    name: str = ""


@dataclass_json
@dataclass
class Tutor:
    id: int
    student: Optional[Student]


@dataclass_json
@dataclass(frozen=True)
class Professor:
    id: int
    name: str


@dataclass_json
@dataclass(frozen=True)
class Course:
    id: int
    name: str
    professor: Professor
    students: Set[Student]


s1 = Student(1, 'student')
s2 = Student(2, 'student')
t = Tutor(id=1, student=None)
p = Professor(1, 'professor')
c = Course(1, 'course', p, {s1})


class TestEncoder:
    def test_student(self):
        assert s1.to_json() == '{"id": 1, "name": "student"}'

    def test_professor(self):
        assert p.to_json() == '{"id": 1, "name": "professor"}'

    def test_course(self):
        assert c.to_json() == '{"id": 1, "name": "course", "professor": {"id": 1, "name": "professor"}, "students": [{"id": 1, "name": "student"}]}'

    def test_students_missing(self):
        s1_anon = Student(1, '')
        s2_anon = Student(2, '')
        one = [s1_anon, s2_anon]
        two = [s2_anon, s1_anon]
        actual = Student.schema().loads('[{"id": 1}, {"id": 2}]', many=True)
        assert actual == one or actual == two


class TestDecoder:
    def test_tutor(self):
        assert Tutor.from_json('{"id": 1}', infer_missing=True) == t



from dataclasses import dataclass

from dataclasses_json import dataclass_json
from datetime import date
import dataclasses_json
import dataclasses_json.cfg


@dataclass_json
@dataclass
class Person:
    name: str


@dataclass_json
@dataclass
class PersonWithBirthday:
    name: str
    birthday: date


class TestGlobalConfig:
    def test_encoder_override(self):
        dataclasses_json.cfg.global_config.encoders[str] = lambda s: s[::-1]
        assert Person("Kobe Bryant").to_json() == '{"name": "tnayrB eboK"}'
        dataclasses_json.cfg.global_config.encoders = {}

    def test_encoder_extension(self):
        dataclasses_json.cfg.global_config.encoders[date] = date.isoformat
        assert PersonWithBirthday("Kobe Bryant", date(1978, 8, 23)).to_json() \
               == '{"name": "Kobe Bryant", "birthday": "1978-08-23"}'
        dataclasses_json.cfg.global_config.encoders = {}



from hypothesis import example


def examples(*args):
    """A variadic `examples` decorator to both supplant stacking of @example
    and support iterables being passed in directly
    """

    def examples_decorator(f):
        g = f
        for arg in args:
            g = example(arg)(g)
        return g

    return examples_decorator



"""
 PEP 0673 that allows self-references has only been added in 3.11
 This suite verifies that we are somewhat able to serde self-referencing classes before 3.11.
"""
import json
from typing import Dict

import pytest

from tests.entities import DataClassWithSelf


class TestSelf:
    @pytest.mark.parametrize("entity, expected", [
        (DataClassWithSelf(id="1", ref=None), {
            "id": "1",
            "ref": None
        }),
        (DataClassWithSelf(id="1", ref=DataClassWithSelf(id="2", ref=None)), {
            "id": "1",
            "ref": { "id": "2", "ref": None }
        })
    ])
    def test_self_serde(self, entity: DataClassWithSelf, expected: Dict):
        assert entity.to_dict() == expected and DataClassWithSelf.from_dict(expected).to_json() == entity.to_json()

    @pytest.mark.parametrize("entity", [
        ({
            "id": "1",
            "ref": {"id": "2", "ref": None}
        }),
        ({
            "id": "1",
            "ref": None
        })
    ])
    def test_self_type(self, entity: Dict):
        obj = DataClassWithSelf.from_json(json.dumps(entity))
        assert isinstance(obj, DataClassWithSelf) and (isinstance(obj.ref, DataClassWithSelf) or isinstance(obj.ref, type(None)))



from dataclasses import dataclass
from typing import Dict, Union, List
import json

from dataclasses_json import dataclass_json


@dataclass_json
@dataclass(frozen=True)
class TestChild:
    some_field: int = None


@dataclass_json
@dataclass(frozen=True)
class TestOtherChild:
    other_field: int = None


@dataclass_json
@dataclass(frozen=True)
class DictUnion:
    d: Dict[str, Union[TestChild, TestOtherChild]]


@dataclass_json
@dataclass(frozen=True)
class ListUnion:
    l: List[Union[TestChild, TestOtherChild]]


@dataclass_json
@dataclass(frozen=True)
class Player:
    name: str


@dataclass_json
@dataclass(frozen=True)
class Team:
    roster: List[Union[int, Player]]
    roster_backup: Dict[int, Union[int, Player]]


class TestCollectionOfUnions:
    def test_dict(self):
        data = {
            'd': {
                'child' : {
                    'some_field' : 1
                },
                'other_child' : {
                    'other_field' : 2
                }
            }
        }
        json_str = json.dumps(data)
        obj = DictUnion.from_json(json_str)

        assert type(obj.d['child']) in (TestChild, TestOtherChild)

    def test_list(self):
        data = {
            'l': [
                {
                    'some_field' : 1
                },
                {
                    'other_field' : 1
                }
            ]
        }
        json_str = json.dumps(data)
        obj = ListUnion.from_json(json_str)

        assert type(obj.l[0]) in (TestChild, TestOtherChild)

    def test_int(self):
        data = {
            'roster': [
                1,
                2,
                3
            ],
            'roster_backup': {
                1: 5,
                2: 3,
                3: 2
            }
        }
        json_str = json.dumps(data)
        obj: Team = Team.from_json(json_str)

        assert type(obj.roster[0]) is int and type(obj.roster_backup[1]) is int

    def test_dataclass(self):
        data = {
            'roster': [
                {
                    'name': 'player1'
                },
                {
                    'name': 'player2'
                },
                {
                    'name': 'player3'
                }
            ],
            'roster_backup': {
                1: {
                    'name': 'player1'
                },
                2: {
                    'name': 'player2'
                },
                3: {
                    'name': 'player3'
                }
            }
        }
        json_str = json.dumps(data)
        obj: Team = Team.from_json(json_str)

        assert type(obj.roster[0]) is Player and type(obj.roster_backup[1]) is Player

    def test_mixed(self):
        data = {
            'roster': [
                {
                    'name': 'player1'
                },
                {
                    'name': 'player2'
                },
                {
                    'name': 'player3'
                }
            ],
            'roster_backup': {
                1: 5,
                2: 3,
                3: 2
            }
        }
        json_str = json.dumps(data)
        obj: Team = Team.from_json(json_str)

        assert type(obj.roster[0]) is Player and type(obj.roster_backup[1]) is int

    def test_mixed_inverse(self):
        data = {
            'roster': [
                1,
                2,
                3
            ],
            'roster_backup': {
                1: {
                    'name': 'player1'
                },
                2: {
                    'name': 'player2'
                },
                3: {
                    'name': 'player3'
                }
            }
        }
        json_str = json.dumps(data)
        obj: Team = Team.from_json(json_str)

        assert type(obj.roster[0]) is int and type(obj.roster_backup[1]) is Player



import json
import logging
import os
# noinspection PyCompatibility
from dataclasses import dataclass
from io import StringIO
from typing import Any, Dict, List, NewType, Optional, Tuple, Union

from mypy.main import main as mypy_main
import pytest

from dataclasses_json import DataClassJsonMixin, CatchAll


@dataclass
class User(DataClassJsonMixin):
    id: str
    name: str = "John"
    ca: CatchAll = None


Filename = NewType('Filename', str)
LineNumber = NewType('LineNumber', int)
ErrorLevel = NewType('ErrorLevel', str)
ErrorMessage = NewType('ErrorMessage', str)


class TestAnnotations:
    u: User = User('ax9ssFxH')
    j: str = u.to_json()
    u2: User = User.from_json(j)
    u2a: User = User.from_json(j.encode())

    jMany = [{"id": "115412", "name": "Peter"},
             {"id": "atxXxGhg", "name": "Parker"}]
    sch = User.schema()
    users1: List[User] = sch.loads(json.dumps(jMany), many=True)
    n: str = users1[1].name
    users2: List[User] = sch.load(jMany, many=True)  # type: ignore
    u3: User = sch.load(jMany[1])
    j2: Dict[str, Any] = sch.dump(u)
    j3: List[Dict[str, Any]] = sch.dump([u2, u3], many=True)
    j4: str = sch.dumps(u2)

    j4_dict: Dict[str, Any] = json.loads(j4)
    u4a: User = User.from_json(j4)
    u4b: User = User.from_dict(j4_dict)

    def filter_errors(self, errors: List[str]) -> List[str]:
        real_errors: List[str] = list()
        current_file = __file__
        current_path = os.path.split(current_file)

        for line in errors:
            line = line.strip()
            if (not line):
                continue

            fn, lno, lvl, msg = self.parse_trace_line(line)
            if (fn is not None):
                _path = os.path.split(fn)
                if (_path[-1] != current_path[-1]):
                    continue

            real_errors.append(line)

        return real_errors

    def parse_trace_line(self, line: str) -> \
            Tuple[Optional[Filename], Optional[LineNumber], Optional[
                ErrorLevel], ErrorMessage]:
        # Define variables
        file_name: Union[str, Filename, None]
        line_no: Union[str, LineNumber, None]
        level: Union[str, ErrorLevel, None]
        msg: Union[str, ErrorMessage, None]

        where, sep, msg = line.partition(': ')
        if (sep):
            file_name, sep, line_no = where.rpartition(':')
            file_name = Filename(file_name)
            if (sep):
                line_no = LineNumber(int(line_no))
            else:
                line_no = None

            level, sep, msg = msg.partition(': ')
            if (sep):
                level = ErrorLevel(level)
            else:
                msg = level
                level = None
        else:
            # Otherwise we get 'Found 1 error in 1 file (checked 1 source file)' as an error
            # due to a mypy error in a different file
            file_name = Filename("") if line.startswith("Found") else None
            line_no = None
            level = None
            msg = line

        msg = ErrorMessage(msg)
        return file_name, line_no, level, msg

    @pytest.mark.skip(reason="mypy_main signature changed")
    def test_type_hints(self):
        text_io = StringIO('')
        try:
            # mypy.main uses sys.stdout for printing
            # We override it to catch error messages
            mypy_main(args=[__file__], stdout=text_io, stderr=text_io, clean_exit=True)
        except SystemExit:
            # mypy.main could return errors found inside other files.
            # filter_errors() will filter out all errors outside this file.
            errors = text_io.getvalue().splitlines()
            errors = self.filter_errors(errors)
        else:
            errors = None

        # To prevent large errors raise error out of try/except
        if (errors):
            logging.error('\n'.join(errors))
            raise AssertionError("Type annotations check failed")



from datetime import datetime, timezone
from dataclasses import dataclass, field
import sys

from marshmallow import fields
import pytest

from dataclasses_json import dataclass_json
from dataclasses_json.mm import _IsoField


@dataclass_json
@dataclass
class DataClassWithDatetime:
    created_at: datetime


if sys.version_info >= (3, 7):
    @dataclass_json
    @dataclass
    class DataClassWithIsoDatetime:
        created_at: datetime = field(
            metadata={'dataclasses_json': {
                'encoder': datetime.isoformat,
                'decoder': datetime.fromisoformat,
                'mm_field': fields.DateTime(format='iso')
            }})


    @dataclass_json
    @dataclass
    class DataClassWithCustomIsoDatetime:
        created_at: datetime = field(
            metadata={'dataclasses_json': {
                'encoder': datetime.isoformat,
                'decoder': datetime.fromisoformat,
                'mm_field': _IsoField()
            }})


class TestTime:
    dt = datetime(2018, 11, 17, 16, 55, 28, 456753, tzinfo=timezone.utc)
    tz = timezone.utc

    ts = dt.timestamp()
    dc_ts_json = f'{{"created_at": {ts}}}'
    dc_ts = DataClassWithDatetime(datetime.fromtimestamp(ts, tz=tz))

    if sys.version_info >= (3, 7):
        iso = dt.isoformat()
        dc_iso_json = f'{{"created_at": "{iso}"}}'
        dc_iso = DataClassWithIsoDatetime(datetime.fromisoformat(iso))

    def test_datetime_encode(self):
        assert (self.dc_ts.to_json() == self.dc_ts_json)

    def test_datetime_decode(self):
        assert (DataClassWithDatetime.from_json(self.dc_ts_json) == self.dc_ts)

    @pytest.mark.skipif(sys.version_info < (3, 7),
                        reason="requires python3.7")
    def test_datetime_override_encode(self):
        assert (self.dc_iso.to_json() == self.dc_iso_json)

    @pytest.mark.skipif(sys.version_info < (3, 7),
                        reason="requires python3.7")
    def test_datetime_override_decode(self):
        assert (DataClassWithIsoDatetime.from_json(
            self.dc_iso_json) == self.dc_iso)

    def test_datetime_schema_encode(self):
        assert (DataClassWithDatetime.schema().dumps(self.dc_ts)
                == self.dc_ts_json)

    def test_datetime_schema_decode(self):
        assert (DataClassWithDatetime.schema().loads(self.dc_ts_json)
                == self.dc_ts)

    @pytest.mark.skipif(sys.version_info < (3, 7),
                        reason="requires python3.7")
    def test_datetime_override_schema_encode(self):
        assert (DataClassWithIsoDatetime.schema().dumps(self.dc_iso)
                == self.dc_iso_json)

    @pytest.mark.skipif(sys.version_info < (3, 7),
                        reason="requires python3.7")
    def test_datetime_override_schema_decode(self):
        iso = DataClassWithIsoDatetime.schema().loads(self.dc_iso_json)
        # FIXME bug in marshmallow currently returns datetime-naive instead of
        # datetime-aware. also seems to drop microseconds?
        # #955
        iso.created_at = iso.created_at.replace(microsecond=456753,
                                                tzinfo=self.tz)
        assert (iso == self.dc_iso)

    @pytest.mark.skipif(sys.version_info < (3, 7),
                        reason="requires python3.7")
    def test_datetime_custom_iso_fieldoverride_schema_encode(self):
        assert (DataClassWithCustomIsoDatetime.schema().dumps(self.dc_iso)
                == self.dc_iso_json)

    @pytest.mark.skipif(sys.version_info < (3, 7),
                        reason="requires python3.7")
    def test_datetime_custom_iso_field_override_schema_decode(self):
        iso = DataClassWithCustomIsoDatetime.schema().loads(self.dc_iso_json)
        assert (iso == DataClassWithCustomIsoDatetime(self.dt))



from dataclasses import dataclass, field
from typing import Any, Dict, List

import pytest
import marshmallow

from dataclasses_json.core import Json
from dataclasses_json.api import dataclass_json, LetterCase, Undefined, DataClassJsonMixin
from dataclasses_json import CatchAll
from dataclasses_json.undefined import UndefinedParameterError


@dataclass_json(undefined=Undefined.INCLUDE)
@dataclass()
class UnknownAPIDump:
    endpoint: str
    data: Dict[str, Any]
    catch_all: CatchAll


@dataclass_json(undefined=Undefined.INCLUDE)
@dataclass()
class UnknownAPIDumpNoCatchAllField:
    endpoint: str
    data: Dict[str, Any]


@dataclass_json(undefined=Undefined.RAISE)
@dataclass()
class WellKnownAPIDump:
    endpoint: str
    data: Dict[str, Any]


@dataclass_json(undefined=Undefined.EXCLUDE)
@dataclass
class DontCareAPIDump:
    endpoint: str
    data: Dict[str, Any]


@pytest.fixture
def valid_response() -> Dict[Any, Json]:
    return {"endpoint": "some_api_endpoint", "data": {"foo": 1, "bar": "2"}}


@pytest.fixture()
def invalid_response(valid_response):
    invalid_response = valid_response.copy()
    invalid_response["undefined_field_name"] = [1, 2, 3]
    return invalid_response


@pytest.fixture()
def invalid_response_camel_case(valid_response):
    invalid_response = valid_response.copy()
    invalid_response["undefinedFieldName"] = [1, 2, 3]
    return invalid_response


@pytest.fixture()
def boss_json():
    boss_json = """
    {
        "minions": [
            {
                "name": "evil minion", 
                "UNKNOWN_PROPERTY" : "value"
            },
            {
                "name": "very evil minion"
            }
        ],
        "UNKNOWN_PROPERTY" : "value"
    }
    """.strip()
    return boss_json


def test_undefined_parameters_catch_all_invalid_back(invalid_response):
    dump = UnknownAPIDump.from_dict(invalid_response)
    inverse_dict = dump.to_dict()
    assert inverse_dict == invalid_response


def test_undefined_parameters_catch_all_valid(valid_response):
    dump = UnknownAPIDump.from_dict(valid_response)
    assert dump.catch_all == {}


def test_undefined_parameters_catch_all_no_field(invalid_response):
    with pytest.raises(UndefinedParameterError):
        UnknownAPIDumpNoCatchAllField.from_dict(invalid_response)


def test_undefined_parameters_catch_all_multiple_fields(invalid_response):
    @dataclass_json(undefined=Undefined.INCLUDE)
    @dataclass()
    class UnknownAPIDumpMultipleCatchAll:
        endpoint: str
        data: Dict[str, Any]
        catch_all: CatchAll
        catch_all2: CatchAll

    with pytest.raises(UndefinedParameterError):
        UnknownAPIDumpMultipleCatchAll.from_dict(invalid_response)


def test_undefined_parameters_catch_all_works_with_letter_case(invalid_response_camel_case):
    @dataclass_json(undefined=Undefined.INCLUDE, letter_case=LetterCase.CAMEL)
    @dataclass()
    class UnknownAPIDumpCamelCase:
        endpoint: str
        data: Dict[str, Any]
        catch_all: CatchAll

    dump = UnknownAPIDumpCamelCase.from_dict(invalid_response_camel_case)
    assert {"undefinedFieldName": [1, 2, 3]} == dump.catch_all
    assert invalid_response_camel_case == dump.to_dict()


def test_undefined_parameters_catch_all_raises_if_initialized_with_catch_all_field_name(valid_response):
    valid_response["catch_all"] = "some-value"
    with pytest.raises(UndefinedParameterError):
        UnknownAPIDump.from_dict(valid_response)


def test_undefined_parameters_catch_all_initialized_with_dict_and_more_unknown(invalid_response):
    invalid_response["catch_all"] = {"someValue": "some-stuff"}
    dump = UnknownAPIDump.from_dict(invalid_response)
    assert dump.catch_all == {"someValue": "some-stuff", "undefined_field_name": [1, 2, 3]}


def test_undefined_parameters_raise_invalid(invalid_response):
    with pytest.raises(UndefinedParameterError):
        WellKnownAPIDump.from_dict(invalid_response)


def test_undefined_parameters_raise_valid(valid_response):
    assert valid_response == WellKnownAPIDump.from_dict(valid_response).to_dict()


def test_undefined_parameters_ignore(valid_response, invalid_response):
    from_valid = DontCareAPIDump.from_dict(valid_response)
    from_invalid = DontCareAPIDump.from_dict(invalid_response)
    assert from_valid == from_invalid


def test_undefined_parameters_ignore_to_dict(invalid_response, valid_response):
    dump = DontCareAPIDump.from_dict(invalid_response)
    dump_dict = dump.to_dict()
    assert valid_response == dump_dict


def test_undefined_parameters_ignore_nested_schema(boss_json):
    @dataclass_json(undefined=Undefined.EXCLUDE)
    @dataclass(frozen=True)
    class Minion:
        name: str

    @dataclass_json(undefined=Undefined.EXCLUDE)
    @dataclass(frozen=True)
    class Boss:
        minions: List[Minion]

    boss = Boss.schema().loads(boss_json)
    assert len(boss.minions) == 2
    assert boss.minions == [Minion(name="evil minion"), Minion(name="very evil minion")]


def test_undefined_parameters_raise_nested_schema(boss_json):
    @dataclass_json(undefined=Undefined.RAISE)
    @dataclass(frozen=True)
    class Minion:
        name: str

    @dataclass_json(undefined=Undefined.EXCLUDE)
    @dataclass(frozen=True)
    class Boss:
        minions: List[Minion]

    with pytest.raises(marshmallow.exceptions.ValidationError):
        Boss.schema().loads(boss_json)


def test_undefined_parameters_catch_all_nested_schema(boss_json):
    @dataclass_json(undefined=Undefined.INCLUDE)
    @dataclass(frozen=True)
    class Minion:
        name: str
        catch_all: CatchAll

    @dataclass_json(undefined=Undefined.INCLUDE)
    @dataclass(frozen=True)
    class Boss:
        minions: List[Minion]
        catch_all: CatchAll

    boss = Boss.schema().loads(boss_json)
    assert {"UNKNOWN_PROPERTY": "value"} == boss.catch_all
    assert {"UNKNOWN_PROPERTY": "value"} == boss.minions[0].catch_all
    assert {} == boss.minions[1].catch_all


def test_undefined_parameters_catch_all_schema_dump(boss_json):
    import json

    @dataclass_json(undefined=Undefined.INCLUDE)
    @dataclass(frozen=True)
    class Minion:
        name: str
        catch_all: CatchAll

    @dataclass_json(undefined=Undefined.INCLUDE)
    @dataclass(frozen=True)
    class Boss:
        minions: List[Minion]
        catch_all: CatchAll

    boss = Boss.schema().loads(boss_json)
    assert json.loads(boss_json) == Boss.schema().dump(boss)
    assert "".join(boss_json.replace('\n', '').split()) == "".join(Boss.schema().dumps(boss).replace('\n', '').split())


def test_undefined_parameters_catch_all_schema_roundtrip(boss_json):
    @dataclass_json(undefined=Undefined.INCLUDE)
    @dataclass(frozen=True)
    class Minion:
        name: str
        catch_all: CatchAll

    @dataclass_json(undefined=Undefined.INCLUDE)
    @dataclass(frozen=True)
    class Boss:
        minions: List[Minion]
        catch_all: CatchAll

    boss1 = Boss.schema().loads(boss_json)
    dumped_s = Boss.schema().dumps(boss1)
    boss2 = Boss.schema().loads(dumped_s)
    assert boss1 == boss2


def test_undefined_parameters_catch_all_ignore_mix_nested_schema(boss_json):
    @dataclass_json(undefined=Undefined.EXCLUDE)
    @dataclass(frozen=True)
    class Minion:
        name: str

    @dataclass_json(undefined=Undefined.INCLUDE)
    @dataclass(frozen=True)
    class Boss:
        minions: List[Minion]
        catch_all: CatchAll

    boss = Boss.schema().loads(boss_json)
    assert Minion(name="evil minion") == boss.minions[0]
    assert Minion(name="very evil minion") == boss.minions[1]
    assert {"UNKNOWN_PROPERTY": "value"} == boss.catch_all


def test_it_works_from_string(invalid_response):
    @dataclass_json(undefined="include")
    @dataclass()
    class UnknownAPIDumpFromString:
        endpoint: str
        data: Dict[str, Any]
        catch_all: CatchAll

    dump = UnknownAPIDumpFromString.from_dict(invalid_response)
    assert {"undefined_field_name": [1, 2, 3]} == dump.catch_all


def test_string_only_accepts_valid_actions():
    with pytest.raises(UndefinedParameterError):
        @dataclass_json(undefined="not sure what this is supposed to do")
        @dataclass()
        class WontWork:
            endpoint: str


def test_undefined_parameters_raises_with_default_argument_and_supplied_catch_all_name(invalid_response):
    @dataclass_json(undefined="include")
    @dataclass()
    class UnknownAPIDumpDefault:
        endpoint: str
        data: Dict[str, Any]
        catch_all: CatchAll = None

    invalid_response["catch_all"] = "this should not happen"
    with pytest.raises(UndefinedParameterError):
        UnknownAPIDumpDefault.from_dict(invalid_response)


def test_undefined_parameters_doesnt_raise_with_default(valid_response, invalid_response):
    @dataclass_json(undefined="include")
    @dataclass()
    class UnknownAPIDumpDefault:
        endpoint: str
        data: Dict[str, Any]
        catch_all: CatchAll = None

    from_valid = UnknownAPIDumpDefault.from_dict(valid_response)
    from_invalid = UnknownAPIDumpDefault.from_dict(invalid_response)
    assert from_valid.catch_all is None
    assert {"undefined_field_name": [1, 2, 3]} == from_invalid.catch_all


def test_undefined_parameters_doesnt_raise_with_default_factory(valid_response, invalid_response):
    @dataclass_json(undefined="include")
    @dataclass()
    class UnknownAPIDumpDefault(DataClassJsonMixin):
        endpoint: str
        data: Dict[str, Any]
        catch_all: CatchAll = field(default_factory=dict)

    from_valid = UnknownAPIDumpDefault.from_dict(valid_response)
    from_invalid = UnknownAPIDumpDefault.from_dict(invalid_response)
    assert from_valid.catch_all == {}
    assert {"undefined_field_name": [1, 2, 3]} == from_invalid.catch_all


def test_undefined_parameters_catch_all_init_valid(valid_response):
    dump = UnknownAPIDump(**valid_response)
    assert dump.catch_all == {}


def test_undefined_parameters_catch_all_init_invalid(invalid_response):
    dump = UnknownAPIDump(**invalid_response)
    assert {"undefined_field_name": [1, 2, 3]} == dump.catch_all


def test_undefined_parameters_catch_all_init_args():
    dump = UnknownAPIDump("some-endpoint", {"some-data": "foo"}, "unknown1", "unknown2", undefined="123")
    assert dump.endpoint == "some-endpoint"
    assert dump.data == {"some-data": "foo"}
    assert dump.catch_all == {'_UNKNOWN0': 'unknown1', '_UNKNOWN1': 'unknown2', "undefined": "123"}


def test_undefined_parameters_catch_all_init_args_kwargs_mixed():
    dump = UnknownAPIDump("some-endpoint", {"some-data": "foo"}, "unknown1", "unknown2", catch_all={"bar": "example"},
                          undefined="123")
    assert dump.endpoint == "some-endpoint"
    assert dump.data == {"some-data": "foo"}
    assert dump.catch_all == {'_UNKNOWN0': 'unknown1', '_UNKNOWN1': 'unknown2', "bar": "example", "undefined": "123"}


def test_undefined_parameters_ignore_init_args():
    dump = DontCareAPIDump("some-endpoint", {"some-data": "foo"}, "unknown1", "unknown2", undefined="123")
    assert dump.endpoint == "some-endpoint"
    assert dump.data == {"some-data": "foo"}


def test_undefined_parameters_ignore_init_invalid(invalid_response, valid_response):
    dump_invalid = DontCareAPIDump(**invalid_response)
    dump_valid = DontCareAPIDump(**valid_response)
    assert dump_valid == dump_invalid


def test_undefined_parameters_raise_init(invalid_response):
    with pytest.raises(TypeError):
        WellKnownAPIDump(**invalid_response)


def test_undefined_parameters_catch_all_default_no_undefined(valid_response):
    @dataclass_json(undefined="include")
    @dataclass()
    class UnknownAPIDumpDefault:
        endpoint: str
        data: Dict[str, Any]
        catch_all: CatchAll = None

    dump = UnknownAPIDumpDefault.from_dict(valid_response)
    assert dump.to_dict() == valid_response


def test_undefined_parameters_catch_all_default_factory_init_converts_factory(valid_response):
    @dataclass_json(undefined="include")
    @dataclass()
    class UnknownAPIDumpDefault:
        endpoint: str
        data: Dict[str, Any]
        catch_all: CatchAll = field(default_factory=dict)

    dump = UnknownAPIDumpDefault(**valid_response)
    assert dump.catch_all == {}


def test_undefined_inheritance():
    @dataclass_json(undefined=Undefined.INCLUDE)
    @dataclass
    class TestInternalConfig:
        options: CatchAll = None
        val: str = "bar"
        val2: int = 0

    @dataclass_json
    @dataclass
    class TestInternalExtendConfig(TestInternalConfig):
        val: str = "baz"

    tie = TestInternalExtendConfig()
    tie.to_dict()




import json
from enum import Enum, IntFlag, auto
from typing import Dict, List
import pytest

from dataclasses import dataclass, field

from dataclasses_json import dataclass_json

from marshmallow.exceptions import ValidationError


class MyEnum(Enum):
    STR1 = "str1"
    STR2 = "str2"
    STR3 = "str3"
    INT1 = 1
    FLOAT1 = 1.23


class MyStrEnum(str, Enum):
    STR1 = "str1"


class MyFlag(IntFlag):
    A = auto()
    B = auto()
    C = auto()


@dataclass_json
@dataclass(frozen=True)
class DataWithEnum:
    name: str
    my_enum: MyEnum = MyEnum.STR3
    enum_list: List[MyEnum] = field(default_factory=list)
    nested: Dict[str, List[MyEnum]] = field(default_factory=dict)


d1 = DataWithEnum('name1', MyEnum.STR1)
d1_json = '{"name": "name1", "my_enum": "str1", "enum_list": [], "nested": {}}'

# Make sure the enum is set to the default value defined by MyEnum
d2_using_default_value = DataWithEnum('name2')
d2_json = '{"name": "name2", "my_enum": "str3", "enum_list": [], "nested": {}}'

d3_int = DataWithEnum('name1', MyEnum.INT1)
d3_int_json = '{"name": "name1", "my_enum": 1, "enum_list": [], "nested": {}}'

d4_float = DataWithEnum('name1', MyEnum.FLOAT1)
d4_float_json = '{"name": "name1", "my_enum": 1.23, "enum_list": [], "nested": {}}'

d5_list = DataWithEnum('name1', MyEnum.STR1, [MyEnum.STR2, MyEnum.STR3], nested={'enum_val': [MyEnum.STR1]})
d5_list_json = '{"name": "name1", "my_enum": "str1", "enum_list": ["str2", "str3"], "nested": {"enum_val": ["str1"]}}'


@dataclass_json
@dataclass(frozen=True)
class DataWithStrEnum:
    my_str_enum: MyStrEnum = MyEnum.STR1


ds = DataWithStrEnum(MyStrEnum.STR1)
ds_json = '{"my_str_enum": "str1"}'


@dataclass_json
@dataclass(frozen=True)
class EnumContainer:
    enum_list: List[MyEnum]
    dict_enum_value: Dict[str, MyEnum]


container_json = '{"enum_list": ["str3", 1], "dict_enum_value": {"key1str": "str1", "key1float": 1.23}}'
container = EnumContainer(
    enum_list=[MyEnum.STR3, MyEnum.INT1],
    dict_enum_value={"key1str": MyEnum.STR1, "key1float": MyEnum.FLOAT1})


@dataclass_json
@dataclass(frozen=True)
class FlagContainer:
    flag_list: List[MyFlag]
    flag: MyFlag

flag_container_json = '{"flag_list": [1, 4], "flag": 1}'
flag_container = FlagContainer(
    flag_list=[MyFlag.A, MyFlag.C],
    flag=MyFlag.A
)


class TestEncoder:
    def test_data_with_enum(self):
        assert d1.to_json() == d1_json, f'Actual: {d1.to_json()}, Expected: {d1_json}'
        assert d3_int.to_json() == d3_int_json, f'Actual: {d3_int.to_json()}, Expected: {d3_int_json}'
        assert d4_float.to_json() == d4_float_json, f'Actual: {d4_float.to_json()}, Expected: {d4_float_json}'

    def test_data_with_str_enum(self):
        assert ds.to_json() == ds_json, f'Actual: {ds.to_json()}, Expected: {ds_json}'

    def test_data_with_enum_default_value(self):
        d2_to_json = d2_using_default_value.to_json()
        assert d2_to_json == d2_json, f"A default value was not included in the JSON encode. " \
                                      f"Expected: {d2_json}, Actual: {d2_to_json}"

    def test_collection_with_enum(self):
        assert container.to_json() == container_json

    def test_enum_with_list(self):
        assert d5_list.to_json() == d5_list_json, f'Actual: {d5_list.to_json()}, Expected: {d5_list_json}'
        assert d5_list.to_dict(encode_json=True) == json.loads(d5_list_json), f'Actual: {d5_list.to_dict()}, Expected: {json.loads(d5_list_json)}'

    def test_enum_flag(self):
        assert flag_container.to_json() == flag_container_json, f'Actual: {flag_container.to_json()}, Expected: {flag_container_json}'
        assert flag_container.to_dict(encode_json=True) == json.loads(flag_container_json), f'Actual: {flag_container.to_dict()}, Expected: {json.loads(flag_container_json)}'


class TestDecoder:
    def test_data_with_enum(self):
        d1_from_json = DataWithEnum.from_json(d1_json)
        assert d1 == d1_from_json
        assert d1_from_json.to_json() == d1_json

        d3_int_from_json = DataWithEnum.from_json(d3_int_json)
        assert d3_int == d3_int_from_json
        assert d3_int_from_json.to_json() == d3_int_json

        d4_float_from_json = DataWithEnum.from_json(d4_float_json)
        assert d4_float == d4_float_from_json
        assert d4_float_from_json.to_json() == d4_float_json

    def test_data_with_str_enum(self):
        ds_from_json = DataWithStrEnum.from_json(ds_json)
        assert ds == ds_from_json
        assert ds_from_json.to_json() == ds_json

    def test_data_with_enum_default_value(self):
        d2_from_json = DataWithEnum.from_json(d2_json)
        assert d2_using_default_value == d2_from_json
        json_from_d2 = d2_from_json.to_json()
        assert json_from_d2 == d2_json, f"A default value was not included in the JSON encode. " \
                                        f"Expected: {d2_json}, Actual: {json_from_d2}"

    def test_collection_with_enum(self):
        container_from_json = EnumContainer.from_json(container_json)
        assert container == container_from_json
        assert container_from_json.to_json() == container_json

    def test_enum_flag(self):
        flag_container_from_json = FlagContainer.from_json(flag_container_json)
        assert flag_container == flag_container_from_json
        assert flag_container_from_json.to_json() == flag_container_json


class TestValidator:
    @pytest.mark.parametrize('enum_value, is_valid', [
        ('str1', True),
        ('str2', True),
        ('str3', True),
        (1, False),
        (1.23, False),
        ('str4', False),
        (2, False),
        (1.24, False),
    ])
    def test_data_with_enum(self, enum_value, is_valid):
        data = '{"name": "myname", "my_enum": "' + str(enum_value) + '"}'
        schema = DataWithEnum.schema()
        res = schema.validate(json.loads(data))
        no_errors = not res
        assert no_errors == is_valid, str(res)

    @pytest.mark.parametrize('enum_value, is_valid', [
        ('str1', True),
        ('str2', False),
    ])
    def test_data_with_str_enum(self, enum_value, is_valid):
        data = '{"my_str_enum": "' + str(enum_value) + '"}'
        schema = DataWithStrEnum.schema()
        res = schema.validate(json.loads(data))
        no_errors = not res
        assert no_errors == is_valid


class TestLoader:
    @pytest.mark.parametrize('json_data, expected_data', [
        (d1_json, d1),
        (d2_json, d2_using_default_value),
        (d3_int_json, d3_int),
        (d4_float_json, d4_float),
    ])
    def test_data_with_enum(self, json_data, expected_data):
        schema = DataWithEnum.schema()
        assert schema.loads(json_data) == expected_data

    def test_data_with_enum_exception(self):
        schema = DataWithEnum.schema()
        with pytest.raises(ValidationError):
            schema.loads('{"name": "name1", "my_enum": "str4"}')

    @pytest.mark.parametrize('json_data, expected_data', [
        (ds_json, ds),
    ])
    def test_data_with_str_enum(self, json_data, expected_data):
        schema = DataWithStrEnum.schema()
        assert schema.loads(json_data) == expected_data

    def test_data_with_str_enum_exception(self):
        schema = DataWithStrEnum.schema()
        with pytest.raises(ValidationError):
            schema.loads('{"my_str_enum": "str2"}')



from collections import deque

from hypothesis.strategies import lists, none, one_of


def deques(elements=None, min_size=0, max_size=None,
           unique_by=None, unique=False):
    return lists(**locals()).map(deque)


def optionals(strategy):
    return one_of(strategy, none())



import pytest
from tests.entities import (DataClassMappingBadDecode,
                            DataClassWithDataClass,
                            DataClassWithList,
                            DataClassWithNestedDictWithTupleKeys,
                            DataClassX,
                            DataClassXs)


class TestEncoder:
    def test_nested_dataclass(self):
        assert (DataClassWithDataClass(DataClassWithList([1])).to_json() ==
                '{"dc_with_list": {"xs": [1]}}')

    def test_nested_list_of_dataclasses(self):
        assert (DataClassXs([DataClassX(0), DataClassX(1)]).to_json() ==
                '{"xs": [{"x": 0}, {"x": 1}]}')


class TestDecoder:
    def test_nested_dataclass(self):
        assert (DataClassWithDataClass.from_json(
            '{"dc_with_list": {"xs": [1]}}') ==
                DataClassWithDataClass(DataClassWithList([1])))

    def test_nested_list_of_dataclasses(self):
        assert (DataClassXs.from_json('{"xs": [{"x": 0}, {"x": 1}]}') ==
                DataClassXs([DataClassX(0), DataClassX(1)]))

    def test_nested_mapping_of_dataclasses(self):
        with pytest.raises(TypeError, match="positional arguments"):
            DataClassMappingBadDecode.from_dict(dict(map=dict(test=dict(id="irrelevant"))))


class TestNested:
    def test_tuple_dict_key(self):
        assert (DataClassWithNestedDictWithTupleKeys.from_dict({'a': {(0,): 2}}) ==
                DataClassWithNestedDictWithTupleKeys(a={(0,): 2}))




import marshmallow
import pytest

from .entities import (DataClassDefaultListStr, DataClassDefaultOptionalList, DataClassList, DataClassOptional,
                       DataClassWithNestedOptional, DataClassWithNestedOptionalAny, DataClassWithNestedAny,
                       DataClassDifferentTypeDecode)
from .test_letter_case import CamelCasePerson, KebabCasePerson, SnakeCasePerson, FieldNamePerson

test_do_list = """[{}, {"children": [{"name": "a"}, {"name": "b"}]}]"""
test_list = '[{"children": [{"name": "a"}, {"name": "b"}]}]'
nested_optional_data = '{"a": {"test": null}}'


class TestSchema:
    def test_default_list_str(self):
        DataClassDefaultListStr.schema().dumps(DataClassDefaultListStr())
        assert True

    def test_default_optional_list(self):
        DataClassDefaultOptionalList.schema().loads(test_do_list, many=True)
        assert True

    def test_list(self):
        DataClassList.schema().loads(test_list, many=True)
        assert True

    def test_optional(self):
        DataClassOptional.schema().loads('{"a": 4, "b": null}')
        assert True

    def test_not_providing_required_field(self):
        with pytest.raises(marshmallow.ValidationError):
            DataClassOptional.schema().loads('{"b": null}')

    def test_letter_case(self):
        for cls in (CamelCasePerson, KebabCasePerson, SnakeCasePerson, FieldNamePerson):
            p = cls('Alice')
            assert p.to_dict() == cls.schema().dump(p)

    def test_nested_optional(self):
        DataClassWithNestedOptional.schema().loads(nested_optional_data)
        assert True

    def test_nested_optional_any(self):
        DataClassWithNestedOptionalAny.schema().loads(nested_optional_data)
        assert True

    def test_nested_any_accepts_optional(self):
        DataClassWithNestedAny.schema().loads(nested_optional_data)
        assert True

    def test_accounts_for_decode(self):
        assert DataClassDifferentTypeDecode.schema().load({'lst': '1,2,3'}) == \
               DataClassDifferentTypeDecode(lst=['1', '2', '3'])



from dataclasses import dataclass
from typing import TypeVar, Generic

from dataclasses_json import DataClassJsonMixin


T = TypeVar('T')
class ContainerCls(Generic[T]):
    value: T
    required: bool

@dataclass
class DataClassWithUnsubscriptedGeneric(DataClassJsonMixin):
    a: ContainerCls

@dataclass
class DataClassWithSubscriptedGeneric(DataClassJsonMixin):
    a: ContainerCls[int]

@dataclass
class DataClassWithParameterizedGeneric(DataClassJsonMixin, Generic[T]):
    a: ContainerCls[T]


class TestUnsupportedGenerics:
    def test_unsubscripted(self):
        j = '{ "a": { "value": "test", "required": true } }'
        dc = DataClassWithUnsubscriptedGeneric.from_json(j)
        assert dc.a == dict(value='test', required=True)
    
    def test_subscripted(self):
        j = '{ "a": { "value": 5, "required": false } }'
        dc = DataClassWithSubscriptedGeneric.from_json(j)
        assert dc.a == dict(value=5, required=False)
    
    def test_parameterized_unsubscripted(self):
        j = '{ "a": { "value": "test", "required": true } }'
        dc = DataClassWithParameterizedGeneric.from_json(j)
        assert dc.a == dict(value='test', required=True)
    
    def test_parameterized_subscripted(self):
        j = '{ "a": { "value": "test", "required": true } }'
        dc = DataClassWithParameterizedGeneric[str].from_json(j)
        assert dc.a == dict(value='test', required=True)
